    <hostdev mode='subsystem' type='pci' managed='yes'>
      <source>
        <address domain='{{ domain }}' bus='{{ bus }}' slot='{{ slot }}' function='{{ function }}'/>
      </source>
    </hostdev>
